
package Bundle::DBD::KingBase;

use strict;
use warnings;
use 5.008001;

our $VERSION = '3.10.12';

1;

__END__

=head1 NAME

Bundle::DBD::KingBase - A bundle to install all DBD::KingBase related modules

=head1 SYNOPSIS

C<perl -MCPAN -e 'install Bundle::DBD::KingBase'>

=head1 CONTENTS

DBI

DBD::KingBase

=head1 DESCRIPTION

This bundle includes all the modules needed for DBD::KingBase (the Perl 
interface to the KingbaseES database system).

=cut

=head1 AUTHOR

kingbase

=cut

